# -*- coding: utf-8 -*-

# Script to automatically verify FOO with multiple votes
# by Jannik Dreier, jannik.dreier@imag.fr , 2011

import subprocess

# defines the maximum number of votes for a voter, e.g. each voter votes between 1 and n times.
n = 4

# generates the voter process.
# instead of using one process and a substitution to determine how many times each voter votes, we will generate different processes, as this is easier to verify for proverif.
def processV(file, n):
  file.write("let processV = \n")
  for k in range(n):
    file.write("  new blinder%d ; new r%d ;\n" % (k, k))
    file.write("  let commitedvote%d = commit(v%d, r%d) in\n" % (k, k, k))
    file.write("  let blindedvote%d = blind(commitedvote%d, blinder%d) in\n" % (k, k, k))
    file.write("  out(ch1, (pk(skv), sign(blindedvote%d, skv)));\n" % k)
  for k in range(n):
    file.write("  in(ch2, m%d);\n" % k)
    file.write("  let result%d = checksign(m%d, pka) in\n" % (k, k))
    file.write("  if result%d = blindedvote%d then\n" % (k, k))
    file.write("  let signedvote%d = unblind(m%d, blinder%d) in\n" % (k, k, k))
  file.write("  sync 1;\n")
  for k in range(n):
    file.write("  out(ch3, (commitedvote%d, signedvote%d));\n" % (k, k))
    file.write("  in(ch4, (l%d, =commitedvote%d, =signedvote%d));\n" % (k, k, k))
  file.write("  sync 2;\n")
  for k in range(n):
    file.write("  out(ch5, (l%d, r%d));\n" % (k, k))
  file.write("  0.\n")
  
# generates the voter under control of the attacker
def processVc1c2(file, n):
  file.write("let processVc1c2 = \n")
  for k in range(n):
    file.write("  new blinder%d; out(c1, blinder%d); new r%d; out(c1, r%d);\n" % (k, k, k, k))
    file.write("  let commitedvote%d = commit(v%d, r%d) in\n" % (k, k, k))
    file.write("  let blindedvote%d = blind(commitedvote%d, blinder%d) in\n" % (k, k, k))
    file.write("  in(c2, y1); out(ch1, y1);\n")
  for k in range(n):
    file.write("  in(ch2, m%d);out(c1, m%d);\n" % (k, k))
    file.write("  let result%d = checksign(m%d, pka) in\n" % (k, k))
    file.write("  in(c2, y2); if y2 = true then\n")
    file.write("  let signedvote%d = unblind(m%d, blinder%d) in\n" % (k, k, k))
  file.write("  sync 1;\n")
  for k in range(n):
    file.write("  in(c2, y3); out(ch3, y3);\n")
    file.write("  in(ch4, k%d); out(c1, k%d);\n" % (k, k))
    file.write("  let (l, commitedvoter, signedvoter) = k%d in\n" % k)
    file.write("  in(c2, y4);if y4 = true then\n")
    file.write("  in(c2, y5);if y4 = true then\n")
  file.write("  sync 2;\n")
  for k in range(n):
    file.write("    in(c2, y6);out(ch5, y6);\n")
  file.write("  0.\n")

def printPI(n, m):
  file = open("foo/%d-%d.piswap" % (n, m), "w")
  file.write("""free ch, ch1, ch2, ch3, ch4, ch5, c1, c2.
free a, b, v.
free ska, skva, skvb, skvc, pka.
free skvaCh, skvbCh, skvcCh.

fun open/2.
fun commit/2.
fun checksign/2.
fun sign/2.
fun blind/2.
fun unblind/2.
fun pk/1.
fun true/0.

equation open(commit(m, r), r) = m.
equation checksign(sign(m, sk), pk(sk)) = m.
equation unblind(blind(m, r), r) =  m.
equation unblind(sign(blind(m, r), sk), r) = sign(m, sk).
""")

  processV(file, n)
  
  file.write("\n")
  
  processVc1c2(file, m)

  file.write("""
  
process
(
(let skv = skva in let v0 = choice[a, b] in processV) |
(let skv = skvb in let v0 = choice[b, a] in processV) |
(let skv = skvc in processVc1c2)
)
""")

  file.close

# iterate over all cases
for l1 in range(1, n):
  for l2 in range(1, n):
    # generate code
    printPI(l1, l2)
    output = open("foo/%d-%d.pi" % (l1, l2), "w")
    # start proswapper
    proswapper = subprocess.Popen(["proswapper", "foo/%d-%d.piswap" % (l1, l2)], stdout=output)
    proswapper.wait()
    output.close()
    # select the correct synchronization process
    output = file("foo/%d-%d.pi" % (l1, l2), "r")
    newlines = []
    for line in output:
      if "| sync_process_0" in line:
        line = "| sync_process_2"
      newlines.append(line)
    output.close()
    outfile = file("foo/%d-%d.pi" % (l1, l2), "w")
    outfile.writelines(newlines)
    outfile.close()
    # start proverif
    outverif = open("foo/%d-%d.verif" % (l1, l2), "w")
    proverif = subprocess.Popen(["proverif", "-in", "pi", "foo/%d-%d.pi" % (l1, l2)], stdout=outverif)
    proverif.wait()
    outverif.close()
    # test for sucess
    outverif = open("foo/%d-%d.verif" % (l1, l2), "r")
    tofind = 'RESULT Observational equivalence is true'
    sucess = False
    for line in outverif:
      if tofind in line:
        sucess = True
    if sucess:
      print "Case %d-%d: OK" % (l1, l2)
    else:
      print "Case %d-%d: KO" % (l1, l2)
    outverif.close()
